module RubyYacht
  # This class models an app that's part of the local system.
  #
  # See RubyYacht::App::DSL for information on configuring the app.
  class App
    # The name of the app
    attr_accessor :name
    
    # The name of the repository where the app's code is stored, relative
    # To the project's repository domain.
    attr_accessor :repository_name
    
    # The type of app this is, which determines what hooks and scripts we run
    # when building it.
    #
    # For instance, this would be `rails` for a Rails app.
    attr_accessor :server_type
    
    # The port that the app listens on.
    attr_accessor :port

    # The name of the database that this app uses as its data store.
    attr_accessor :database_name

    # The project that includes this app.
    def project
      RubyYacht.configuration.projects.find { |project| project.apps.include?(self) }
    end

    # This method gets the name of the image / container that this app will run
    # in.
    def container_name
      project = self.project
      if project.system_prefix == self.name
        return "#{project.system_prefix}"
      else
        return "#{project.system_prefix}-#{self.name}"
      end
    end

    # This method gets the database that the app connects to.
    #
    # ### Returns
    # RubyYacht::Database
    def database
      project.databases.find { |database| database.name == database_name }
    end
    
    # This class provide's the DSL for configuring an app.
    #
    # Inside of the project configuration, you can call `app [name]` to add an
    # app, and then pass it a block to configure the app, which will allow you
    # to call these DSL methods.
    class DSL
      include RubyYacht::DSL::Base
      extend RubyYacht::DSL::Base::ClassMethods
      
      # This initializer starts the DSL for the app.
      #
      # ### Parameters
      #
      # * **server_type: Symbol**   The type of server this is.
      # * **name: String**          The name of the app.
      def initialize(server_type, name)
        @server_type = server_type
        @name = name.to_sym
        load_custom_attributes
      end
      
      add_attribute :name
      
      ##
      # :method: repository_name
      # You can call `repository_name 'foo/bar'` to set the app's `repository_name`.
      add_attribute :repository_name, nil, false
      
      ##
      # :method: database_name
      # You can call `database 'foo'` to tell the app to use the database server
      # named `foo`.
      add_attribute :database_name, nil, false
      
      ##
      # :method: server_type
      # You can call `server_type :foo` to set the app's `server_type`.
      add_attribute :server_type
      
      ##
      # :method: port
      # You can call `port 3000` to set the app's `port`. It defaults to 8080.
      add_attribute :port, 8080

      creates_object App
      
      # This method checks that all of the required attributes have been set on
      # the object.
      #
      # If they haven't, this will raise an exception.
      #
      # It also checks that the server type has been defined in the
      # configuration.
      def check_required_attributes
        super
        check_server_type @server_type, :app
      end
    end
  end
end