require 'yaml'
require_relative 'project'
require_relative 'hook'
require_relative 'server_type'

module RubyYacht
  # This class stores the configuration for the system.
  #
  # For more information on the configuration DSL, see RubyYacht::Configuration::DSL.
  class Configuration
    # This initializer creates an empty configuration.
    def initialize
      self.clear
    end

    # This method erases all the configuration.
    def clear
      @projects = []
      @hooks = []
      @server_types = []
      @avoid_docker_machine = false
      @local_config = {}
    end
    
    # The projects that are part of this system.
    # Each entry is a RubyYacht::Project
    attr_accessor :projects

    # The hooks to customize the build and run processes.
    # Each entry is a RubyYacht::Hook
    attr_accessor :hooks

    # The app types that we can support.
    # Each entry is a Symbol.
    attr_accessor :server_types

    # Whether we should avoid using docker-machine even if it is installed.
    #
    # If docker-machine is not installed, the value of this flag will be
    # ignored.
    attr_accessor :disable_docker_machine

    # The configuration that has been loaded from a local YAML file.
    attr_accessor :local_config

    # This method pulls up the hooks that we have defined.
    #
    # ### Parameters
    #
    # * **server: App/Database**:   The server we are fetching hooks for.
    # * **event_type: Symbol**:     The event that we are fetching hooks for.
    #
    # ### Returns
    #
    # The matching hooks. This will be an Array where each item is a Hook.
    def fetch_hooks(server, event_type)
      case server
      when App
        app_types = [server.server_type,:all]
        database_types = [:all]
        database = server.database
        database_types << database.server_type if database
        web_server_types = [:all] + server.project.web_servers.map(&:server_type).uniq
        container_type = :app
      when Database
        database_types = [server.server_type, :all]
        app_types = [:all]
        app_types += server.apps.map(&:server_type)
        web_server_types = [:all] + server.project.web_servers.map(&:server_type).uniq
        container_type = :database
      when WebServer
        app_types = [:all] + server.project.apps.map(&:server_type).uniq
        database_types = [:all] + server.project.databases.map(&:server_type).uniq
        web_server_types = [:all, server.server_type]
        container_type = :web
      else
        return []
      end
      self.hooks.select do |hook|
        hook.event_type == event_type &&
          app_types.include?(hook.app_server_type) &&
          database_types.include?(hook.database_server_type) &&
          web_server_types.include?(hook.web_server_type) &&
          hook.container_type == container_type
      end
    end

    # This method finds an server type by name.
    #
    # ### Parameters
    #
    # * **name: Symbol**    The name of the server type to return.
    #
    # ### Returns
    #
    # The RubyYacht::ServerType with that name.
    def find_server_type(name)
      self.server_types.find { |type| type.name == name }
    end
    
    # This method provides a DSL for top-level configuration.
    class DSL
      include RubyYacht::DSL::Base
      extend RubyYacht::DSL::Base::ClassMethods

      # This initializer creates an empty configuration DSL.
      def initialize
        self.load_custom_attributes
        @hook_options = {}
        @local_config = {}
      end
      
      ##
      # :method: project
      #
      # This method adds a project to the configuration.
      #
      # This takes the project name as its first argument. It also takes a block
      # which you can use to configure the project, using
      # RubyYacht::Project::DSL.
      add_object_list :project, RubyYacht::Project::DSL

      ##
      # :method: hook
      #
      # This method adds a hook to the configuration.
      #
      # This takes hook's event_time and event_type as its arguments.  It also
      # takes a block which you can use to configure the hook, using
      # RubyYacht::Hook::DSL.
      add_object_list :hook, RubyYacht::Hook::DSL

      ##
      # :method: server_type
      #
      # This method adds a server type to the configuration.
      #
      # This takes type's name as its argument. It also takes a block which you
      # can use to configure the type, using
      # RubyYacht::ServerType::DSL.
      add_object_list :server_type, RubyYacht::ServerType::DSL

      # The default options for the hooks we're defining in the current block.
      attr_accessor :hook_options

      # The configuration that has been loaded from a local YAML file.
      attr_accessor :local_config

      # This method sets default attributes for a group of hooks.
      #
      # Any hooks that you create in the associated block will have the included
      # options set on them automatically. You will also be able to override
      # these options in the configuration blocks for individual hooks.
      #
      # ### Parameters
      #
      # * **options: Hash**     The fields to set on the hooks.
      # * **block**             A block for adding the hooks.
      def add_hooks(options = {}, &block)
        old_options = hook_options.dup
        self.hook_options = options
        block.call
        self.hook_options = old_options
      end

      # This method loads config from a local YAML file.
      #
      # The contents of the file will be stored in the global configuration's
      # `local_config` field. From there, it can be loaded into another DSL
      # object by calling `copy_local_config`.
      #
      # ### Parameters
      #
      # * **path: String**    The full path to the config file.
      def add_local_config(path)
        new_config = YAML.load_file(path)
        @local_config = @local_config.merge(new_config)
      end

      # This method adds a before hook.
      #
      # ### Parameters
      #
      # * **event_type: Symbol**    The event type for the new hook.
      # * **block**                 A block for configuring the hook. You can 
      #                             call the RubyYacht::Hook::DSL methods in
      #                             this block.
      def before(event_type, &block)
        add_hook :before, event_type, &block
      end

      # This method adds an after hook.
      #
      # ### Parameters
      #
      # * **event_type: Symbol**    The event type for the new hook.
      # * **block**                 A block for configuring the hook. You can 
      #                             call the RubyYacht::Hook::DSL methods in
      #                             this block.
      def after(event_type, &block)
        add_hook :after, event_type, &block
      end

      # This method adds a during hook.
      #
      # This hook will be run while event is happening, and provides the core
      # logic for the event.
      #
      # ### Parameters
      #
      # * **event_type: Symbol**    The event type for the new hook.
      # * **block**                 A block for configuring the hook. You can 
      #                             call the RubyYacht::Hook::DSL methods in
      #                             this block.
      def during(event_type, &block)
        add_hook :during, event_type, &block
      end
      
      creates_object Configuration, %w(projects hooks server_types local_config)

      private

      def add_hook(event_time, event_type, &block)
        hook_options = self.hook_options
        hook event_time, event_type do
          hook_options.each do |key, value|
            public_send(key, value)
          end
          instance_eval(&block)
        end
      end
    end
  end
  
  # This method adds configuration for the system.
  #
  # If you pass a block to this method, it will be evaluated using the methods
  # in RubyYacht::Configuration::DSL.
  #
  # Any projects you define in that block will be added to the current list
  # of projects.
  def self.configure(&block)
    new_configuration = Configuration::DSL.new.run(block).create_object
    %w(projects hooks).each do |field|
      self.configuration.send("#{field}=", self.configuration.send(field) + new_configuration.send(field))
    end

    self.configuration.local_config = self.configuration.local_config.merge(new_configuration.local_config)

    new_configuration.server_types.each do |type|
      if self.configuration.server_types.any? { |existing| existing.name == type.name }
        raise "Server type already registered: #{type.name}"
      end
      self.configuration.server_types << type
    end
  end
  
  # This method gets the current configuration for the system.
  def self.configuration
    @configuration ||= Configuration.new
  end
end