module RubyYacht
  # This class stores the configuration for a project's database server.
  #
  # You can configure a database with RubyYacht::Database::DSL
  class Database
    # The host that the database server is accessed on.
    attr_accessor :host
    
    # The name of the database
    attr_accessor :name
    
    # The type of database server we're using.
    attr_accessor :server_type
    
    # The username we use to connect to the database
    attr_accessor :username
    
    # The password we use to connect to the database
    attr_accessor :password
    
    # The port the database server should listen on.
    attr_accessor :port

    # The database-specific part of the container name for this database.
    attr_accessor :container_label
    
    # Whether the database is running on the same machine as the app servers.
    def local?
      host == 'localhost'
    end

    # The project that contains this database.
    def project
      RubyYacht.configuration.projects.find { |project| project.databases.include?(self) }
    end

    # The apps that connect to this database.
    def apps
      project.apps.select { |app| app.database_name == self.name }
    end

    # This method gets the name of the container and image for this database.
    #
    # This will be the project's prefix followed by the database's container
    # label.
    def container_name
      "#{project.system_prefix}-#{container_label}"
    end

    # This class provides a DSL for configuring a database.
    #
    # You can access this DSL by calling `database` within `RubyYacht::Project::DSL`.
    # It will create a RubyYacht::Database
    class DSL
      include RubyYacht::DSL::Base
      extend RubyYacht::DSL::Base::ClassMethods
      
      # This initializer starts the DSL for the database.
      #
      # ### Parameters
      #
      # * **server_type: Symbol**   The type of database this is.
      # * **name: String**          The name of the database.
      def initialize(server_type, name)
        @server_type = server_type
        @name = name
        load_custom_attributes
      end
      
      ##
      # :method: host
      # You can call `host 'db.test.com'` to set the database's host.
      add_attribute :host
      
      #
      # :method: name
      # You can call `name` 'my_database'` to tell the apps to use the database
      # named `my_database`.
      add_attribute :name
      
      #
      # :method: server_type
      # You can call `server_type` :sqlite` to tell the apps to use the database
      # server type sqlite.
      add_attribute :server_type
      
      ##
      # :method: username
      # You can call `username 'db-user'` to tell the apps to connect to the
      # database under the name `db-user`.
      add_attribute :username
      
      ##
      # :method: password
      # You can call `password 'testpass'` to tell the apps to connect to the
      # password `testpass`.
      add_attribute :password
      
      ##
      # :method: port
      # You can call `port 1234` to tell the database server to listen on port
      # 1234.
      add_attribute :port
      
      ##
      # :method: container_label
      # You can call `container_label 'mysql'` to give this database a container
      # name that is (project)-mysql.
      add_attribute :container_label, :database
      
      creates_object Database
      
      # This method checks that all of the required attributes have been set on
      # the object.
      #
      # If they haven't, this will raise an exception.
      #
      # It also checks that the server type has been defined in the
      # configuration.
      def check_required_attributes
        super
        check_server_type @server_type, :database
      end
    end
  end
end