module RubyYacht
  # This module contains types for defining our configuration DSL.
  module DSL
    # This module provides the core functionality for providing configuration
    # DSLs.
    #
    # Classes that provide DSLs should include this module and extend the
    # ClassMethods module.
    module Base
      # This module provides class methods for defining DSLs.
      module ClassMethods
        # The attributes that must be provided on this DSL type.
        def required_attributes
          @required_attributes ||= []
        end
        
        # The attributes that this DSL type can accept.
        def all_attributes
          @all_attributes ||= []
        end
        
        # The default values that should be set on new DSL objects.
        def default_values
          @default_values ||= {}
        end

        # The attributes that we copy from this type to the real type the DSL
        # creates.        
        def copied_attributes
          @copied_attributes
        end
        
        # The real type that this DSL creates.
        def created_type
          @created_type
        end
        
        # This method adds an attribute to the DSL.
        #
        # This will define a method on the DSL class, and register the attribute
        # with the class.
        #
        # You generally won't need to call this yourself, but it provides a
        # shared implementation for the rest of the attribute-definition
        # methods.
        #
        # ### Parameters
        #
        # * **method_name: Symbol**       The name of the method to create.
        # * **attribute_name: Symbol**    The name of the instance variable that
        #                                 the method sets.
        # * **default: Object**           The default value that should be set
        #                                 for this attribute.
        # * **required: Bool**            Whether the user should be required to
        #                                 provide a value for this attribute.
        # * **processing_block: Proc**    The block that we should invoke when
        #                                 the method is called, to set the value
        #                                 of the attribute.
        def add_generic_attribute(method_name, attribute_name, default = nil, required = true, &processing_block)
          define_method(method_name, &processing_block)
          if default != nil
            default_values[attribute_name] = default
          end
          if required
            required_attributes << attribute_name
          end
          all_attributes << attribute_name
        end
        
        # This method adds a standard attribute to the DSL.
        #
        # This will create a method that you can invoke with a single parameter,
        # which provides the new value for the attribute.
        #
        # ### Parameters
        #
        # * **name: Symbol**      The name of the attribute, which will also be
        #                         the name of the method.
        # * **default: Object**   The default value for the attribute.
        # * **required: Bool**    Whether we should require the user to provide
        #                         a value for the attribute.
        def add_attribute(name, default = nil, required = true)
          add_generic_attribute name, name, default, required do |value|
            instance_variable_set("@#{name}", value)
          end
        end
        
        # This method adds an attribute to the DSL for managing a list.
        #
        # This will create a method that you can invoke with a single parameter,
        # which will be added to the list.
        #
        # ### Parameters
        #
        # * **name: Symbol**    The name of the method. The name of the
        #                       attribute will be this name, followed by an s.
        def add_list(name)
          add_generic_attribute name, "#{name}s".to_sym, [], false do |value|
            variable_name = "@#{name}s"
            list = instance_variable_get(variable_name)
            list << value
            instance_variable_set(variable_name, list)
          end
        end
        
        # This method adds an attribute to the DSL for managing a Boolean flag.
        #
        # This will create a method that you can call to change the value from
        # true to false.
        #
        # The value will always be false by default.
        #
        # ### Parameters
        #
        # * **name: Symbol**    The name of the method, which will also be the
        #                       name of the attribute.
        def add_boolean(name)
          add_generic_attribute name, name, false, false do
            instance_variable_set("@#{name}", true)
          end
        end

        # This method adds an object from another DSL to this DSL.
        #
        # This will create a method that you can call with a block, which will
        # allow you to use the other DSL object. You can also give it arbitrary
        # other arguments, which will be given to the initializer for the DSL.
        #
        # ### Parameters
        #
        # * **name: Symbol**    The name of the method, which will also be the
        #                       name of the attribute.
        # * **type: Class**     The class type that provides the DSL when they
        #                       call this method.
        # * **options: Hash**   Other options about the DSL. Right now the only
        #                       option is *required*, which indicates whether
        #                       they must define this configuraiton.
        def add_object(name, type, options = {})
          unless options.has_key?(:required)
            options[:required] = true
          end
          add_generic_attribute name, name, nil, options[:required] do |*args, &config_block|
            variable_name = "@#{name}"
            object_config = type.new(*args)
            value = object_config.run(config_block).create_object
            instance_variable_set(variable_name, value)
          end
        end
        
        # This method adds a list of objects from another DSL to this DSL.
        #
        # This will create a method that you can call with a block to create the
        # other object and add it to the list. You can also give it arbitrary
        # other arguments, which will be given to the initializer for the DSL.
        #
        # ### Parameters
        #
        # * **name: Symbol**    The name of the method. The attribute name will
        #                       be this, followed by an "s"
        # * **type: Class**     The class type that provides the DSL when they
        #                       call this method.
        def add_object_list(name, type)
          add_generic_attribute name, "#{name}s".to_sym, [], false do |*args, &config_block|
            variable_name = "@#{name}s"
            list = instance_variable_get(variable_name)
            object_config = type.new(*args)
            object_config.run(config_block)
            value = object_config.create_object
            list << value
            instance_variable_set(variable_name, list)
          end
        end
        
        # This method specifies that this DSL class creates a real object of
        # a certain type.
        #
        # ### Parameters
        #
        # * **type: Class**         The type that this DSL creates.
        # * **attributes: Array**   The attributes that we copy from the DSL
        #                           to the related object. If this is not
        #                           provided, we will copy all of the
        #                           attributes that this DSL has defined so far.
        def creates_object(type, attributes = nil)
          attributes ||= self.all_attributes
          @copied_attributes = attributes
          @created_type = type
        end

        # This method gets the attribute that server types can use to specify
        # custom attributes for this DSL type.
        #
        # The default is `server_attributes`.
        def custom_attribute_method
          :server_attributes
        end
      end

      # This initializer creates a new DSL instance.
      def initialize
        load_custom_attributes
      end

      # This method adds attributes to this DSL instance's singleton DSL, based
      # on the attributes defined by the app types that have been loaded.
      def load_custom_attributes
        full_class = self.class
        self.singleton_class.instance_eval do
          @copied_attributes = (full_class.copied_attributes || []).dup
          @all_attributes = full_class.all_attributes.dup
          @required_attributes = full_class.required_attributes.dup
          @default_values = full_class.default_values.dup
        end

        if self.class.created_type
          RubyYacht.configuration.server_types.each do |server_type|
            next if @server_type && server_type.name != @server_type
            if server_type.respond_to?(self.class.custom_attribute_method)
              attributes = server_type.send(self.class.custom_attribute_method)
              attributes.each do |attribute|
                attribute = attribute.merge(name: "#{server_type.name}_#{attribute[:name]}".to_sym)
                load_custom_attribute(attribute)
              end
            end
          end
        end
      end

      private

      # This method defines the fields for a custom attribute on the DSL.
      #
      # ### Parameters
      #
      # * **attribute: Hash**   The settings on the attribute. The `name`
      #                         setting must already include the app type's
      #                         prefix.
      def load_custom_attribute(attribute)
        if attribute.has_key?(:required)
          required = attribute[:required]
        else
          required = true
        end
        name = attribute[:name]
        self.singleton_class.add_attribute name, attribute[:default], required
        self.singleton_class.copied_attributes << name
        self.class.created_type.instance_eval do
          attr_accessor name
        end
      end

      public
      
      # This method runs a block with DSL methods.
      #
      # The block will reun in the instance context of the DSL object, so you
      # can call the DSL methods directly.
      #
      # It will also copy the default values into the attributes on the DSL
      # object.
      #
      # ### Parameters
      #
      # * **block: Proc**   The block to run.
      def run(block)
        defaults = self.singleton_class.default_values || {}
        defaults.each do |name, value|
          copy =
            case value
            when TrueClass, FalseClass, Fixnum, Symbol
              value
            else
              value.dup
            end
          instance_variable_set("@#{name}", copy)
        end
        if block
          instance_eval(&block)
        end
        
        type = RubyYacht.configuration.find_server_type(@server_type)
        if type
          type.server_defaults.each do |server_defaults|
            server_defaults.each do |key, value|
              if instance_variable_get("@#{key}") == nil
                instance_variable_set("@#{key}", value)
              end
            end
          end
        end
        self
      end

      # This method copies fields from the local config to this DSL.
      #
      # The local config must have been loaded in a previous configuration block
      # using the `add_local_config` method.
      #
      # The keys in the local config file should all be symbols, as should the
      # `fields`. If they are not, this method may not be able to find the
      # config entries.
      #
      # Any fields that are not present in the local config will be set to nil.
      #
      # ### Parameters
      #
      # * **fields: [Symbol]**    The fields to load.
      # * **from: String**        A dot-separated key path to the location in
      #                           the config file that contains the dictionary
      #                           that we are pulling the fields from.
      def copy_local_config(*fields, from:nil)
        dictionary = RubyYacht.configuration.local_config
        if from
          from.split('.').each do |key|
            dictionary = dictionary[key.to_sym]
          end
        end
        fields.each do |field|
          unless self.singleton_class.all_attributes.include?(field)
            raise "Undefined field in #{self.class.name}: #{field}"
          end
          instance_variable_set("@#{field}", dictionary[field])
        end
      end
      
      # This method checks that all of the required attributes have been set on
      # the object.
      #
      # If they haven't, this will raise an exception.
      def check_required_attributes
        attributes = self.singleton_class.required_attributes || []
        attributes.each do |name|
          value = instance_variable_get("@#{name}")
          if value == nil || value == ''
            raise "Missing required attribute #{name} for #{self.class}"
          end
        end
      end

      # This method checks that a server type has been registered.
      #
      # If it has not, this will raise an exception.
      #
      # ### Parameters
      #
      # * **server_type: Symbol**     The server type that has been set on this
      #                               DSL instance.
      # * **container_type: Symbol**    The container type that the server type
      #                                 must be registered for.
      def check_server_type(server_type, container_type)
        result = RubyYacht.configuration.find_server_type(server_type)
        unless result && result.container_type == container_type
          raise "#{self.class.name} has invalid #{container_type} server type `#{server_type}`"
        end
      end
      
      # This method creates an object with the attributes from this instance.
      #
      # We will check that all the required attributes have been set first, and
      # will raise an exception if they haven't.
      #
      # The created type, and the attributes we set on it, will be defined by
      # the call to the class method `creates_object`.
      #
      # ### Returns
      # This will return the newly created object.
      def create_object
        check_required_attributes
        object = self.class.created_type.new
        self.singleton_class.copied_attributes.each do |name|
          object.send("#{name}=", instance_variable_get("@#{name}"))
        end
        object
      end
    end
  end
end