module RubyYacht
  # This class provides a hook that a user or plugin can provide to customize
  # the app images.
  #
  # You can configure this with RubyYacht::Hook::DSL
  class Hook
    # The permitted values for the `event_time` attribute.
    EVENT_TIMES = [:before, :after, :during]

    # The permitted values for the `event_type` attribute.
    EVENT_TYPES = [:startup, :build_checkout, :install_libraries]

    # When this hook should take effect, relative to the event.
    attr_accessor :event_time

    # The type of event that this hook is attached to.
    attr_accessor :event_type

    # The type of app server that this hook applies to.
    attr_accessor :app_server_type

    # The type of database server that this hook applies to.
    attr_accessor :database_server_type

    # The type of web server that this hook applies to.
    attr_accessor :web_server_type

    # The type of container that this hook applies to.
    attr_accessor :container_type

    # The behaviors that we will run for this hook.
    attr_accessor :behaviors

    # This class provides a DSL for configuring hooks.
    class DSL
      include RubyYacht::DSL::Base
      extend RubyYacht::DSL::Base::ClassMethods

      # This initializer creates a new Hook DSL.
      #
      # ### Parameters
      #
      # * **event_time: Symbol**      The time for the hook, relative to the
      #                               event.
      # * **event_type: Symbol**      The type of event the hook is attached to.
      def initialize(event_time, event_type)
        @event_time = event_time
        @event_type = event_type
        @behaviors = []
        load_custom_attributes
      end

      ##
      # :method: app_server_type
      #
      # You can call `app_server_type :rails` to signify that this hook applies
      # to rails apps.
      add_attribute :app_server_type, :all

      ##
      # :method: database_server_type
      #
      # You can call `database_server_type :mysql` to signify that this hook
      # applies to MySQL databases.
      add_attribute :database_server_type, :all

      ##
      # :method: web_server_type
      #
      # You can call `web_server_type :nginx` to signify that this hook
      # applies to nginx web servers.
      add_attribute :web_server_type, :all

      ##
      # :method: container_type
      #
      # You can call `container_type :app` to signify that this hook applies to
      # app images.
      add_attribute :container_type

      ##
      # :method: script_folder
      #
      # You can call `script_folder './scripts'` to signify that the scripts for
      # this hook are found in the `./scripts` folder.
      add_attribute :script_folder, '.'

      # This method adds a behavior to run a command when the hook is run.
      #
      # ### Parameters
      #
      # * **cmd: String**   The command to run.
      def command(cmd)
        @behaviors << CommandBehavior.new(cmd)
      end

      # This method adds a behavior to copy a file to the image when the hook is
      # run.
      #
      # ### Parameters
      #
      # * **path: String**    The path to the file, relative to the
      #                       `script_folder`.
      def copy_file(path)
        @behaviors << CopyFileBehavior.new("#{@script_folder}/#{path}")
      end

      # This method adds a behavior to run a script.
      #
      # The script will be understood to be in the script folder that was given
      # when the DSL was created.
      #
      # ### Parameters
      #
      # * **interpreter: Symbol**     The interpreter that should process the
      #                               script. e.g. ruby or bash
      # * **name: String**            The filename of the script.
      def run_script(interpreter, name)
        copy_file name
        destination = "/var/docker/#{name}"
        command "#{interpreter} #{destination}"
      end

      # This method adds a behavior to set an environment variable.
      #
      # The variable's value can be defined literally, by passing a value as the
      # second parameter to this method, or at build-time, by passing a block.
      #
      # ### Parameters
      #
      # * **name: String**    The name of the environment variable.
      # * **value: String**   A literal value to set for the environment
      #                       variable.
      # * **block: Proc**     A block for providing the environment variable.
      def set_environment_variable(name, value=nil, &block)
        if value
          value_proc = Proc.new { value }
        else
          value_proc = block
        end
        @behaviors << EnvironmentVariableBehavior.new(name, value_proc)
      end
      
      # This method checks that all of the required attributes have been set on
      # the object.
      #
      # If they haven't, this will raise an exception.
      #
      # This will also check that the app type is valid.
      def check_required_attributes
        super
        if @app_server_type != :all
          check_server_type @app_server_type, :app
        end
        if @database_server_type != :all
          check_server_type @database_server_type, :database
        end
        if @web_server_type != :all
          check_server_type @web_server_type, :web
        end
      end

      creates_object RubyYacht::Hook, %w(event_time event_type app_server_type
        database_server_type web_server_type container_type behaviors)
    end

    # This class is the base class for behaviors in hooks.
    class Behavior
      # The command that should be run in a Dockerfile for this behavior.
      def dockerfile_command
        ''
      end

      # The command that should be run in a shell script for this behavior.
      def shell_command
        ''
      end

      # This context in which the command is being generated.
      attr_accessor :context
    end

    # This class provides a behavior for running a command in the image.
    class CommandBehavior < Behavior
      # This initializer creates the behavior.
      #
      # ### Parameters
      #
      # * **command: String**   The command to run.
      def initialize(command)
        @command = command
      end

      # The command that should be run in a Dockerfile for this behavior. 
      def dockerfile_command
        "RUN #{@command}"
      end

      # The command that should be run in a shell script for this behavior.
      def shell_command
        @command
      end
    end

    # This class provides a behavior for copying a file from the host machine
    # to the image.
    class CopyFileBehavior < Behavior
      # This initialize creates the behavior.
      #
      # ### Parameters
      #
      # * **file_path: String**   The path to the file, relative to the root of
      #                           the project.
      def initialize(file_path)
        @file_path = file_path
      end

      # The command that should be run in a Dockerfile for this behavior.
      def dockerfile_command
        "COPY #{File.basename(file_path)} /var/docker/"
      end

      # :method: file_path
      # The path to the file that this behavior copies.
      attr_reader :file_path
    end

    # This class provides a behavior for setting an environment variable in the
    # image.
    class EnvironmentVariableBehavior < Behavior
      # This initializer creates the behavior.
      #
      # ### Parameters
      #
      # * **name: String**    The name of the environment variable to create.
      # * **proc: Proc**      The proc for generating the value for the
      #                       environment variable.
      def initialize(name, proc)
        @name = name
        @proc = proc
        @context = self
      end

      # The value for the environment variable.
      def value
        context.instance_eval(&@proc)
      end

      # The command that should be run in a Dockerfile for this behavior.
      def dockerfile_command
        value = self.value
        return nil unless value
        "ENV #{@name} #{value}"
      end

      # The command that should be run in a shell script for this behavior.
      def shell_command
        value = self.value
        return nil unless value
        "export #{@name}=\"#{value}\""
      end
    end
  end
end