require_relative 'dns_server'
require_relative 'app'
require_relative 'database'
require_relative 'web_server'

module RubyYacht
  # This class provides a configuration for a project in our system.
  #
  # A project is a family of apps that communicate with each other and
  # share a database.
  #
  # You can configure this with RubyYacht::Project::DSL
  class Project
    # The name of the project.
    attr_accessor :name
    
    # The prefix that we use before all of the images and containers for the
    # project.
    attr_accessor :system_prefix
    
    # The hostname for the repository that contains the code for the apps.
    attr_accessor :repository
    
    # The protocol that we use to check out the repository.
    #
    # This can be `ssh`, `http`, or `https`.
    attr_accessor :repository_protocol
    
    # Whether we should check out the code on the host machine, rather than
    # keeping it entirely inside the container.
    attr_accessor :check_out_locally
    
    # The configuration for the databases. Each entry is a RubyYacht::Database.
    attr_accessor :databases
    
    # The configuration for the apps. Each entry is a RubyYacht::App.
    attr_accessor :apps

    # The configuration for the web servers. Each entry is a
    # RubyYacht::WebServer.
    attr_accessor :web_servers
    
    # The configuration for the DNS for the apps. This is a
    # RubyYacht::DnsServer.
    attr_accessor :dns_server

    # The name of the primary app for the project.
    #
    # This app will respond to requests to the bare domain with no app
    # subdomain. It will also respond under its normal subdomain.
    #
    # If this is not provided, this will create a special index page for the
    # bare domain, with links to each app.
    attr_accessor :primary_app
    
    # This class provides a DSL for configuring a project.
    #
    # You can access this inside RubyYacht::Configuration::DSL by calling
    # `project`. It will create a RubyYacht::Project.
    class DSL
      include RubyYacht::DSL::Base
      extend RubyYacht::DSL::Base::ClassMethods
      
      # This initializer creates a new project.
      #
      # ### Parameters
      #
      # * *name: String* The name of the project
      def initialize(name)
        @name = name.to_sym
        load_custom_attributes
        load_app_initializers
      end
      
      add_attribute :name
      
      ##
      # :method: system_prefix
      # You can call `system_prefix 'my-apps'` to set the projects
      # system_prefix` to `my-apps`
      add_attribute :system_prefix
      
      ##
      # :method: repository
      # You can call `repository 'github.com'` to set the project's `repository`
      # to `github.com`
      add_attribute :repository
      
      ##
      # :method: repository_protocol
      # You can call `repository_protocol :https` to tell the scripts to check
      # out the repositories over HTTPS rather than SSH.
      add_attribute :repository_protocol, :ssh

      #
      # :method: primary_app
      # You can call `primary_app :app1` to tell the project to use `app1` as
      # the app under the bare domain.
      add_attribute :primary_app, nil, false
      
      ##
      # :method: check_out_locally
      #
      # You can call `check_out_locally` to set the project's
      # `check_out_locally` flag to `true`. The default is false.
      add_boolean :check_out_locally
      
      ##
      # :method: app
      #
      # You can call `app 'mars'` to add an app called `mars`. This takes a
      # block for configuring the app, which allows you to call methods in
      # RubyYacht::App::DSL
      add_object_list :app, RubyYacht::App::DSL
      
      ##
      # :method: database
      #
      # You can call `database` to configure a database for the project.
      # This takes a block for configuring the database, which allows you to
      # call methods in RubyYacht::Database::DSL
      add_object_list :database, RubyYacht::Database::DSL
      
      ##
      # :method: web_server
      #
      # You can call `web_server` to configure a web server for the project.
      # This takes a block for configuring the server, which allows you to
      # call methods in RubyYacht::WebServer::DSL
      add_object_list :web_server, RubyYacht::WebServer::DSL
      
      ##
      # :method: dns_server
      #
      # You can call `dns_server` to configure the DNS for the project.
      # This takes a block for configuring the DNS , which allows you to
      # call methods in RubyYacht::DnsServer::DSL
      #
      # This is optional.
      add_object :dns_server, RubyYacht::DnsServer::DSL, required: false
      
      creates_object Project

      # This method loads the special DSL methods for defining apps of a certain
      # type.
      #
      # For the app type with the name `foo`, this will create a `foo_app`
      # method on the DSL that builds an app with the type `foo`.
      def load_app_initializers
        RubyYacht.configuration.server_types.each do |server_type|
          main_method = server_type.container_type.to_s
          main_method += "_server" if server_type.container_type == :web
          method_name = "#{server_type.name}_#{main_method}"
          define_singleton_method method_name do |*args, &block|
            self.public_send(
              main_method,
              server_type.name,
              *args,
              &block
            ) 
          end
        end
      end

      # This method gets the attribute that server types can use to specify
      # custom attributes for this DSL type.
      #
      # This DSL uses `project_attributes`.
      def self.custom_attribute_method
        :project_attributes
      end
    end
  end
end