module RubyYacht
  # This class represents a type of server that the user can configure.
  #
  # An server type corresponds to a major server framework, like Ruby on Rails.
  # Server types are defined by plugins, which also provide the logic for
  # installing and running the server.
  #
  # You can configure this with RubyYacht::ServerType::DSL
  class ServerType
    # The name of the type.
    attr_accessor :name

    # The type of container that this should be applied to.
    attr_accessor :container_type

    # The attributes that we define on the project DSL once this server type has
    # been loaded.
    #
    # Each entry will be a hash with a key for `name`, and optionally a key for
    # `default` and `required`. These values will be given to the
    # `add_attribute` method in the project DSL.
    #
    # The attribute names will be prefixed with the server type's name. For
    # instance, if the `rails` server type provides an `environment` attribute,
    # it will be called `rails_environment` on the project's DSL. This prevents
    # conflicts with attributes from other plugins.
    attr_accessor :project_attributes
    
    # The attributes that we define on the server DSL once this server type has
    # been loaded.
    #
    # Each entry will be a hash with a key for `name`, and optionally a key for
    # `default` and `required`. These values will be given to the
    # `add_attribute` method in the project DSL.
    #
    # The attributes will only be defined on servers with this server type.
    #
    # The attribute names will be prefixed with the server type's name. For
    # instance, if the `rails` server type provides an `environment` attribute,
    # it will be called `rails_environment` on the server's DSL. This prevents
    # conflicts with attributes from other plugins.
    attr_accessor :server_attributes
    
    # The default values for the servers with this server type.
    #
    # Each entry will be a hash where the keys are the names of attributes, and
    # the values are the default values.
    attr_accessor :server_defaults

    # The docker image that we use as the source for the server images.
    attr_accessor :baseline_image

    # This class provides a DSL for configuring a RubyYacht::ServerType.
    class DSL
      include RubyYacht::DSL::Base
      extend RubyYacht::DSL::Base::ClassMethods

      # This initiailzer creates the server type DSL.
      #
      # ### Parameters
      #
      # * **name: Symbol**    The name of the server type.
      def initialize(name)
        @name = name
        load_custom_attributes
      end

      add_attribute :name

      #
      # :method: container_type
      # You can call `container_type :app` to say that this server type is for
      # apps. Other acceptable values are `database` and `web`.
      add_attribute :container_type

      #
      # :method: baseline_image
      # You can call `baseline_image 'ubuntu'` to use ubuntu as the source
      # image for the server containers for this server type.
      add_attribute :baseline_image

      #
      # :method: project_attribute
      # You can call `project_attribute name: :environment, default: 'staging'`
      # to add an attribute to the project DSL.
      add_list :project_attribute

      #
      # :method: server_attribute
      # You can call `server_attribute name: :environment, default: 'staging'`
      # to add an attribute to the server DSL.
      add_list :server_attribute

      #
      # :method: server_default
      # You can call `server_default port: 3306` to specify that the servers
      # should have a default value of 3306 for the `port` attribute.
      add_list :server_default

      creates_object RubyYacht::ServerType, [:name, :baseline_image,
        :container_type, :project_attributes, :server_attributes,
        :server_defaults]
    end
  end
end