module RubyYacht
  # This class models a web server that's part of the local system.
  #
  # See RubyYacht::WebServer::DSL for information on configuring the server.
  class WebServer
    # The name of the server
    attr_accessor :name

    # The domain that this web server listens on.
    attr_accessor :domain

    # The port that this web server listens on.
    attr_accessor :port

    # The name of the web server plugin for this server.
    attr_accessor :server_type

    # The project that includes this web server.
    def project
      RubyYacht.configuration.projects.find { |project| project.web_servers.include?(self) }
    end

    # This method gets the name of the image / container that this server will
    # run in.
    def container_name
      project = self.project
      return "#{project.system_prefix}-#{self.name}"
    end
    
    # This class provide's the DSL for configuring a web server.
    #
    # Inside of the project configuration, you can call `web_server [name]` to
    # add a web server, and then pass it a block to configure the server, which
    # will allow you to call these DSL methods.
    #
    # You can also call `web_server` with no argument, which will give it the
    # name `web`.
    class DSL
      include RubyYacht::DSL::Base
      extend RubyYacht::DSL::Base::ClassMethods
      
      # This initializer starts the DSL for the server.
      #
      # ### Parameters
      #
      # * **server_type: Symbol**   The type of server this is.
      # * **name: String**          The name of the server.
      def initialize(server_type, name=:web)
        @server_type = server_type
        @name = name.to_sym
        load_custom_attributes
      end
      
      add_attribute :name
      
      ##
      # :method: domain
      # You can call `domain 'test.com'` to tell the web server to create
      # subdomains under test.com.
      add_attribute :domain
      
      ##
      # :method: server_type
      # You can call `server_type :foo` to set the app's `server_type`.
      add_attribute :server_type
      
      ##
      # :method: port
      # You can call `port 8080` to set the server's `port`. It defaults to 80.
      add_attribute :port, 80

      creates_object WebServer
      
      # This method checks that all of the required attributes have been set on
      # the object.
      #
      # If they haven't, this will raise an exception.
      #
      # It also checks that the server type has been defined in the
      # configuration.
      def check_required_attributes
        super
        check_server_type @server_type, :web
      end
    end
  end
end