require 'fileutils'

module RubyYacht::Runner
  # This command provides help information about other commands.
  class CreateNewProject < Command
    # The name of the command.
    def self.command; 'new'; end

    # The short description of the command.
    def self.description
      'Create a new ruby-yacht project'
    end

    # The project that we are creating.
    attr_accessor :project

    # The directory where we should place the config.
    attr_accessor :directory

    # This method gets the command-line options for the command.
    def option_parser
      OptionParser.new do |options|
        options.banner = "Usage: #{Command.short_script_name} new [PROJECT] [DIRECTORY]\n\n#{self.class.description}"
      end
    end

    # This method extracts arguments from the command line.
    #
    # ### Parameters
    #
    # * **arguments: Array**    The command line arguments.
    def parse_positional_arguments(arguments)
      self.project = arguments.shift
      self.directory = arguments.shift
    end 

    # This method runs the logic for the command.
    def run
      if project.nil?
        log 'You must provide a project'
        log "Run #{Command.short_script_name} help new for more information"
        return false
      end

      if directory.nil?
        log 'You must provide a directory'
        log "Run #{Command.short_script_name} help new for more information"
        return false
      end

      FileUtils.mkdir_p(directory)
      File.open(File.join(directory, 'run.rb'), 'w') do |file|
        write_config_file(file)
      end
      log "Your project has been created in #{directory}/run.rb."
      log "You can go to #{directory} and run `ruby run.rb build` to create your docker containers."
      log "Your initial app will be created in the #{project} directory."
      true
    end

    private

    # This method writes the config file for the new project to a file.
    #
    # ### Parameters
    #
    # * **file: IO**    The file to write to.
    def write_config_file(file)
      file.puts "require 'ruby_yacht'"
      file.puts ''
      file.puts 'RubyYacht.configure do'
      file.puts "  project :#{project} do"
      file.puts "    system_prefix           :#{project}"
      file.puts "    repository              'github.com'"
      file.puts "    rails_secret_key_base   'testkey'"
      file.puts "    check_out_locally"
      file.puts ""
      file.puts "    rails_app :#{project}"
      file.puts ""
      file.puts "    primary_app :#{project}"
      file.puts ""
      file.puts "    nginx_web_server do"
      file.puts "      domain '#{project}.test.com'"
      file.puts "    end"
      file.puts "  end"
      file.puts "end"
      file.puts ""
      file.puts "RubyYacht::Runner.run"
    end
  end
end